# flake8: noqa

from skale.contracts.base_contract import BaseContract, transaction_method

from skale.contracts.manager import Manager
from skale.contracts.contract_manager import ContractManager
from skale.contracts.token import Token
from skale.contracts.groups import Groups
from skale.contracts.constants_holder import ConstantsHolder

from skale.contracts.schains import SChains
from skale.contracts.schains_internal import SChainsInternal
from skale.contracts.nodes import Nodes
from skale.contracts.monitors import Monitors

from skale.contracts.delegation.delegation_controller import DelegationController
from skale.contracts.delegation.validator_service import ValidatorService
from skale.contracts.delegation.token_state import TokenState
from skale.contracts.delegation.distributor import Distributor
from skale.contracts.delegation.slashing_table import SlashingTable

from skale.contracts.dkg import DKG

from skale.contracts.test.time_helpers_with_debug import TimeHelpersWithDebug



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

import logging

from web3 import Web3
from web3.middleware import geth_poa_middleware

import skale.contracts as contracts
from skale.wallets import BaseWallet
from skale.contracts_info import get_base_contracts_info, get_debug_contracts_info
from skale.utils.helper import get_abi
from skale.utils.web3_utils import get_provider
from skale.utils.exceptions import InvalidWalletError, EmptyWalletError, IncompatibleAbiError

logger = logging.getLogger(__name__)


class EmptyPrivateKey(Exception):
    pass


def spawn_skale_lib(skale):
    return Skale(skale._endpoint, skale._abi_filepath, skale.wallet)


class Skale:
    def __init__(self, endpoint, abi_filepath, wallet=None, provider_timeout=30):
        logger.info(f'Init skale-py, connecting to {endpoint}')
        provider = get_provider(endpoint, timeout=provider_timeout)
        self._abi_filepath = abi_filepath
        self._endpoint = endpoint
        self.web3 = Web3(provider)
        self.web3.middleware_onion.inject(
            geth_poa_middleware, layer=0)  # todo: may cause issues
        self.__contracts = {}
        if wallet:
            self.wallet = wallet
        self.__init_contracts(get_abi(abi_filepath))

    @property
    def gas_price(self):
        return self.web3.eth.gasPrice * 5 // 4

    @property
    def wallet(self):
        if not self._wallet:
            raise EmptyWalletError('No wallet provided')
        return self._wallet

    @wallet.setter
    def wallet(self, wallet):
        if issubclass(type(wallet), BaseWallet):
            self._wallet = wallet
        else:
            raise InvalidWalletError(f'Wrong wallet class: {type(wallet).__name__}. \
                                       Must be one of the BaseWallet subclasses')

    def __is_debug_contracts(self, abi):
        return abi.get('time_helpers_with_debug_address', None)

    def __init_contracts(self, abi):
        self.add_lib_contract('contract_manager',
                              contracts.ContractManager, abi)
        self.__init_contracts_from_info(abi, get_base_contracts_info())
        if self.__is_debug_contracts(abi):
            logger.info('Debug contracts found in ABI file')
            self.__init_contracts_from_info(abi, get_debug_contracts_info())

    def __init_contracts_from_info(self, abi, contracts_info):
        for name in contracts_info:
            info = contracts_info[name]
            if info.upgradeable:
                self.init_upgradeable_contract(info, abi)
            else:
                self.add_lib_contract(info.name, info.contract_class, abi)

    def init_upgradeable_contract(self, contract_info, abi):
        address = self.get_contract_address(contract_info.contract_name)
        self.add_lib_contract(contract_info.name, contract_info.contract_class,
                              abi, address)

    def add_lib_contract(self, name, contract_class, abi, contract_address=None):
        address = contract_address or self.get_contract_address_by_name(
            abi, name)
        logger.info(f'Fetching abi for {name}, address {address}')
        contract_abi = self.get_contract_abi_by_name(abi, name)
        self.add_contract(name, contract_class(
            self, name, address, contract_abi))

    def get_contract_address_by_name(self, abi, name):
        return self.get_abi_key(abi, self.get_address_key_name(name))

    def get_contract_abi_by_name(self, abi, name):
        return self.get_abi_key(abi, self.get_abi_key_name(name))

    def get_abi_key_name(self, contract_name):
        # todo: tmp fix for inconsistent contract names
        if contract_name == 'manager' or contract_name == 'token':
            return f'skale_{contract_name}_abi'
        if contract_name == 'dkg':
            return 'skale_d_k_g_abi'
        return f'{contract_name}_abi'

    def get_address_key_name(self, contract_name):
        if contract_name == 'manager' or contract_name == 'token':
            return f'skale_{contract_name}_address'
        if contract_name == 'dkg':
            return 'skale_d_k_g_address'
        return f'{contract_name}_address'

    def get_abi_key(self, abi, key):
        try:
            return abi[key]
        except KeyError:
            raise IncompatibleAbiError(key)

    def add_contract(self, name, skale_contract):
        logger.debug(f'Init contract: {name}')
        self.__contracts[name] = skale_contract

    def get_contract_address(self, name):
        return self.contract_manager.get_contract_address(name)

    def get_contract_by_name(self, name):
        return self.__contracts[name]

    def __getattr__(self, name):
        if name not in self.__contracts:
            return None
        return self.get_contract_by_name(name)



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.
""" Schains.sol functions """

import functools
from Crypto.Hash import keccak

from skale.contracts import BaseContract
from skale.utils.helper import format_fields


FIELDS = [
    'name', 'owner', 'indexInOwnerList', 'partOfNode', 'lifetime', 'startDate', 'startBlock',
    'deposit', 'index', 'chainId'
]


class SChains(BaseContract):
    """Wrapper for some of the Schains.sol functions"""

    @property
    @functools.lru_cache()
    def schains_internal(self):
        return self.skale.get_contract_by_name('schains_internal')

    @format_fields(FIELDS)
    def get(self, id_):
        res = self.schains_internal.get_raw(id_)
        hash_obj = keccak.new(data=res[0].encode("utf8"), digest_bits=256)
        hash_str = "0x" + hash_obj.hexdigest()[:13]
        res.append(hash_str)
        return res

    @format_fields(FIELDS)
    def get_by_name(self, name):
        id_ = self.name_to_id(name)
        res = self.schains_internal.get_raw(id_)
        hash_obj = keccak.new(data=res[0].encode("utf8"), digest_bits=256)
        hash_str = "0x" + hash_obj.hexdigest()[:13]
        res.append(hash_str)
        return res

    def get_schains_for_owner(self, account):
        schains = []
        list_size = self.schains_internal.get_schain_list_size(account)

        for i in range(0, list_size):
            id_ = self.schains_internal.get_schain_id_by_index_for_owner(account, i)
            schain = self.get(id_)
            schains.append(schain)
        return schains

    def get_schains_for_node(self, node_id):
        schains = []
        schain_ids = self.schains_internal.get_schain_ids_for_node(node_id)
        for schain_id in schain_ids:
            schain = self.get(schain_id)
            schain['active'] = True if self.schain_active(schain) else False
            schains.append(schain)
        return schains

    def name_to_id(self, name):
        keccak_hash = keccak.new(data=name.encode("utf8"), digest_bits=256)
        return keccak_hash.hexdigest()

    def get_last_rotation_id(self, schain_name):
        rotation_data = self.schains_internal.get_rotation(schain_name)
        return rotation_data['rotation_id']

    def schain_active(self, schain):
        if schain['name'] != '' and \
                schain['owner'] != '0x0000000000000000000000000000000000000000':
            return True

    def get_schain_price(self, index_of_type, lifetime):
        return self.contract.functions.getSchainPrice(index_of_type,
                                                      lifetime).call()



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.
""" SKALE token operations """

from skale.contracts import BaseContract, transaction_method
from skale.utils.constants import GAS


class Token(BaseContract):
    @transaction_method(gas_limit=GAS['token_transfer'])
    def transfer(self, address, value):
        return self.contract.functions.send(address, value, b'')

    def get_balance(self, address):
        return self.contract.functions.balanceOf(address).call()

    @transaction_method(gas_limit=GAS['token_transfer'])
    def add_authorized(self, address, wallet):  # pragma: no cover
        return self.contract.functions.addAuthorized(address)

    def get_and_update_slashed_amount(self, address: str) -> int:
        return self.contract.functions.getAndUpdateSlashedAmount(address).call()



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

from web3 import Web3

from skale.dataclasses import CurrentNodeInfo, SchainNodeInfo
from skale.schain_config.ports_allocation import get_schain_base_port_on_node
from skale.schain_config import PRECOMPILED_IMA_CONTRACTS
from skale.utils.helper import ip_from_bytes, decapitalize
from skale.utils.web3_utils import public_key_to_address
from skale.schain_config.base_config import update_base_config


def generate_schain_info(schain, schain_nodes,
                         custom_schain_config_fields={},
                         snapshot_interval_ms=None,
                         empty_block_interval_ms=None):
    schain_info = {
        'schainID': 1,  # todo: remove this later (should be removed from the skaled first)
        'schainName': schain['name'],
        'schainOwner': schain['owner'],
        'nodes': schain_nodes
    }
    schain_info.update(custom_schain_config_fields)
    if snapshot_interval_ms:
        schain_info['snapshotIntervalMs'] = snapshot_interval_ms
    if empty_block_interval_ms:
        schain_info['emptyBlockIntervalMs'] = empty_block_interval_ms
    return schain_info


def get_nodes_for_schain(skale, name):
    nodes = []
    ids = skale.schains_internal.get_node_ids_for_schain(name)
    for id_ in ids:
        node = skale.nodes.get(id_)
        node['id'] = id_
        nodes.append(node)
    return nodes


def get_nodes_for_schain_config(skale, name):
    nodes_info = []
    nodes = get_nodes_for_schain(skale, name)
    for i, node in enumerate(nodes, 1):
        schains_on_node = skale.schains.get_schains_for_node(node['id'])
        base_port = get_schain_base_port_on_node(schains_on_node, name, node['port'])

        node_info = SchainNodeInfo(
            node_name=node['name'],
            node_id=node['id'],
            base_port=base_port,

            schain_index=i,
            ip=ip_from_bytes(node['ip']),
            public_key=node['publicKey'],
            public_ip=ip_from_bytes(node['publicIP']),
            owner=public_key_to_address(node['publicKey'])
        ).to_config()
        nodes_info.append(node_info)
    return nodes_info


def get_contract_address_from_ima_data(ima_data, name):
    return ima_data[f'{name}_address']


def generate_permitted_ima_contracts_info(ima_data):
    permitted_contracts = {}
    for name in PRECOMPILED_IMA_CONTRACTS:
        contract_filename = PRECOMPILED_IMA_CONTRACTS[name]['filename']
        permitted_contracts[contract_filename] = get_contract_address_from_ima_data(ima_data, name)
    return permitted_contracts


def generate_mp_authorized_callers(ima_data, schain_owner, schain_nodes):
    mp_authorized_callers = {}
    for name in PRECOMPILED_IMA_CONTRACTS:
        address = get_contract_address_from_ima_data(ima_data, name)
        mp_authorized_callers[address] = 1
    for node in schain_nodes:
        acc_fx = Web3.toChecksumAddress(node['owner'])
        if not mp_authorized_callers.get(acc_fx, None):
            mp_authorized_callers[acc_fx] = 1
    schain_owner_fx = Web3.toChecksumAddress(schain_owner)
    if not mp_authorized_callers.get(schain_owner_fx, None):
        mp_authorized_callers[schain_owner_fx] = 1
    return mp_authorized_callers


def generate_ima_contracts_addresses(ima_data):
    ima_contracts_addresses = {}
    for name in PRECOMPILED_IMA_CONTRACTS:
        contract_filename = PRECOMPILED_IMA_CONTRACTS[name]['filename']
        address = get_contract_address_from_ima_data(ima_data, name)
        ima_contracts_addresses[decapitalize(contract_filename)] = address
    return ima_contracts_addresses


def generate_schain_contract_settings(ima_data, schain_owner, schain_nodes):
    # todo: refactor
    permitted_contracts = generate_permitted_ima_contracts_info(ima_data)
    mp_authorized_callers = generate_mp_authorized_callers(ima_data, schain_owner, schain_nodes)
    ima_contracts_addresses = generate_ima_contracts_addresses(ima_data)
    return {
        'common': {
            'enableContractLogMessages': False
        },
        'IMA': {
            'ownerAddress': schain_owner,
            'variables': {
                'LockAndDataForSchain': {
                    'permitted': permitted_contracts
                },
                'MessageProxyForSchain': {
                    'mapAuthorizedCallers': mp_authorized_callers
                }
            },
            **ima_contracts_addresses
        }
    }


def generate_schain_config(base_config, node_info, schain_info, schain_contract_settings={}):
    return {
        **base_config,
        'skaleConfig': {
            'contractSettings': schain_contract_settings,
            'nodeInfo': node_info,
            'sChain': schain_info,
        }
    }


def generate_skale_schain_config(skale, schain_name, node_id, base_config=None, ima_mainnet=None,
                                 ima_mp_schain=None, ima_mp_mainnet=None, wallets=None,
                                 ima_data=None, rotate_after_block=64,
                                 empty_block_interval_ms=None, snapshot_interval_ms=None,
                                 schain_log_level='info', schain_log_level_config='info',
                                 custom_schain_config_fields={}):
    node = skale.nodes.get(node_id)
    schain = skale.schains.get_by_name(schain_name)

    schains_on_node = skale.schains.get_schains_for_node(node_id)
    schain_base_port_on_node = get_schain_base_port_on_node(schains_on_node, schain_name,
                                                            node['port'])
    schain_nodes = get_nodes_for_schain_config(skale, schain_name)

    node_info = CurrentNodeInfo(
        node_name=node['name'],
        node_id=node_id,
        base_port=schain_base_port_on_node,
        bind_ip=ip_from_bytes(node['ip']),
        ima_mainnet=ima_mainnet,
        ima_mp_schain=ima_mp_schain,
        ima_mp_mainnet=ima_mp_mainnet,
        wallets=wallets,
        rotate_after_block=rotate_after_block,
        schain_log_level=schain_log_level,
        schain_log_level_config=schain_log_level_config
    ).to_config()

    schain_info = generate_schain_info(
        schain=schain,
        schain_nodes=schain_nodes,
        empty_block_interval_ms=empty_block_interval_ms,
        snapshot_interval_ms=snapshot_interval_ms,
        custom_schain_config_fields=custom_schain_config_fields
    )

    if base_config:
        update_base_config(base_config, schain, schain_nodes, ima_data=ima_data)
    else:
        base_config = {}
    if ima_data:
        schain_contract_settings = generate_schain_contract_settings(ima_data, schain['owner'],
                                                                     schain_nodes)
    else:
        schain_contract_settings = {}
    return generate_schain_config(
        base_config=base_config,
        node_info=node_info,
        schain_info=schain_info,
        schain_contract_settings=schain_contract_settings
    )



#   -*- coding: utf-8 -*-
#
#   This file is part of SKALE.py
#
#   Copyright (C) 2019-Present SKALE Labs
#
#   SKALE.py is free software: you can redistribute it and/or modify
#   it under the terms of the GNU Affero General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   SKALE.py is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU Affero General Public License for more details.
#
#   You should have received a copy of the GNU Affero General Public License
#   along with SKALE.py.  If not, see <https://www.gnu.org/licenses/>.

from skale.utils.contracts_provision import (
    D_VALIDATOR_ID, D_VALIDATOR_MIN_DEL, D_DELEGATION_PERIOD, D_DELEGATION_INFO,
    D_VALIDATOR_NAME, D_VALIDATOR_DESC, D_VALIDATOR_FEE, DEFAULT_NODE_NAME, SECOND_NODE_NAME,
    DEFAULT_SCHAIN_NAME
)
from skale.utils.contracts_provision.utils import (
    generate_random_node_data, generate_random_schain_data
)


def _skip_evm_time(web3, seconds) -> int:
    """For test purposes only, works only with ganache node"""
    res = web3.provider.make_request("evm_increaseTime", [seconds])
    web3.provider.make_request("evm_mine", [])
    return res['result']


def cleanup_nodes_schains(skale):
    print('Cleanup nodes and schains')
    for schain_id in skale.schains_internal.get_all_schains_ids():
        schain_data = skale.schains.get(schain_id)
        schain_name = schain_data.get('name', None)
        if schain_name is not None:
            skale.manager.delete_schain(schain_name, wait_for=True)
    for node_id in skale.nodes.get_active_node_ids():
        skale.manager.node_exit(node_id, wait_for=True)


def validator_exist(skale):
    return skale.validator_service.number_of_validators() > 0


def setup_validator(skale):
    """Create and activate a validator"""
    if not validator_exist(skale):
        create_validator(skale)
        enable_validator(skale)
    else:
        print('Skipping default validator creation')
    set_test_msr(skale)

    delegate_to_validator(skale)
    delegations = skale.delegation_controller.get_all_delegations_by_validator(D_VALIDATOR_ID)
    accept_pending_delegation(skale, delegations[-1]['id'])


def link_address_to_validator(skale):
    print('Linking address to validator')
    signature = skale.validator_service.get_link_node_signature(D_VALIDATOR_ID)
    tx_res = skale.validator_service.link_node_address(
        node_address=skale.wallet.address,
        signature=signature,
        wait_for=True
    )
    tx_res.raise_for_status()


def skip_delegation_delay(skale, delegation_id):
    print(f'Activating delegation with ID {delegation_id}')
    skale.token_state._skip_transition_delay(
        delegation_id,
        wait_for=True
    )


def accept_pending_delegation(skale, delegation_id):
    print(f'Accepting delegation with ID: {delegation_id}')
    skale.delegation_controller.accept_pending_delegation(
        delegation_id=delegation_id,
        wait_for=True
    )


def get_test_delegation_amount(skale):
    msr = skale.constants_holder.msr()
    return msr * 30


def set_test_msr(skale):
    skale.constants_holder._set_msr(
        new_msr=D_VALIDATOR_MIN_DEL,
        wait_for=True
    )


def delegate_to_validator(skale):
    print(f'Delegating tokens to validator ID: {D_VALIDATOR_ID}')
    skale.delegation_controller.delegate(
        validator_id=D_VALIDATOR_ID,
        amount=get_test_delegation_amount(skale),
        delegation_period=D_DELEGATION_PERIOD,
        info=D_DELEGATION_INFO,
        wait_for=True
    )


def enable_validator(skale):
    print(f'Enabling validator ID: {D_VALIDATOR_ID}')
    skale.validator_service._enable_validator(D_VALIDATOR_ID, wait_for=True)


def create_validator(skale):
    print('Creating default validator')
    skale.validator_service.register_validator(
        name=D_VALIDATOR_NAME,
        description=D_VALIDATOR_DESC,
        fee_rate=D_VALIDATOR_FEE,
        min_delegation_amount=D_VALIDATOR_MIN_DEL,
        wait_for=True
    )


def create_nodes(skale):
    # create couple of nodes
    print('Creating two nodes')
    node_names = [DEFAULT_NODE_NAME, SECOND_NODE_NAME]
    for name in node_names:
        ip, public_ip, port, _ = generate_random_node_data()
        skale.manager.create_node(ip, port, name, public_ip, wait_for=True)


def create_schain(skale):
    print('Creating schain')
    # create 1 s-chain
    type_of_nodes, lifetime_seconds, _ = generate_random_schain_data()
    price_in_wei = skale.schains.get_schain_price(type_of_nodes,
                                                  lifetime_seconds)

    skale.manager.create_schain(
        lifetime_seconds,
        type_of_nodes,
        price_in_wei,
        DEFAULT_SCHAIN_NAME,
        wait_for=True
    )
